-- Constants
local VERSION = "1.0 alpha"

local ELK_PANEL_SID = "urn:micasaverde-com:serviceId:ElkAlarmPanel1"
local ALARM_PARTITION_SID = "urn:micasaverde-com:serviceId:AlarmPartition2"
local SECURITY_SID = "urn:micasaverde-com:serviceId:SecuritySensor1"
local DEBUG_MODE = true

local TASK_ERROR = 2
local TASK_ERROR_PERM = -2
local TASK_SUCCESS = 4
local TASK_BUSY = 1

-- Globals
local g_remoteArmEnabled = false
local g_remoteDisarmEnabled = false
local g_remotePanicEnabled = false

local g_alarmDevice
local g_isStartupMode = true
local g_taskHandle = -1

local g_childDevices = {
    partitions = {},
    zones = {},
}

local log = luup.log


---- Generic Utilities ----

local function debug (text)
    if (DEBUG_MODE == true) then
        log (text)
    end
end


local function task (text, mode)
    log (string.format ("task '%s'", text))
    if (mode == TASK_ERROR_PERM) then
        g_taskHandle = luup.task (text, TASK_ERROR, "DSCAlarmPanel", g_taskHandle)
    else
        g_taskHandle = luup.task (text, mode, "DSCAlarmPanel", g_taskHandle)

        -- Clear the previous error, since they're all transient
        if (mode ~= TASK_SUCCESS) then
            luup.call_delay ("clearTask", 30, "", false)
        end
    end
end

-- Has to be "non-local" in order for MiOS to call it :(
function clearTask()
    task ("Clearing...", TASK_SUCCESS)
end

-- Concatenates the given strings.
local function strCat (s, ...)
    if (select ("#", ...) == 0) then
        return s
    end
    s = (s == nil) and "" or tostring (s)

    return s .. strCat (...)
end


function table:contains (item)
    for _, v in pairs (self) do
        if v == item then
            return true
        end
    end

    return false
end

-- Trims leading and trailing spaces.
local function trimSpaces (s)
    return s:gsub("^%s*", ""):gsub("%s*$", "")
end


local function rpad(s, len, ch)
    -- TODO: Not strictly correct, since "ch" isn"t guaranteed to be one char,
    --       and the length could end up negative.
    --
    local ch = (ch or " ")
    return s .. ch:rep(len - s:len())
end


local function setMiOSAttribute(name, value, device)
    local setAttr = luup.attr_set

    --
    -- Only call the method if it exists in the MiOS Version we're currently running...
    --
    if (setAttr ~= nil) then
        setAttr(name, value, device)
    else
        debug("setMiOSAttribute: Function luup.attr_set not present in MiOS unit")
    end
end


local function setDSCPanelState(variable, data)
    luup.variable_set(ELK_PANEL_SID, variable, data, g_alarmDevice)
end


local function setPartitionState(partition, variable, data)
    local partitionDevice = g_childDevices.partitions[partition]

    if (partitionDevice ~= nil) then
        luup.variable_set(ALARM_PARTITION_SID, variable, data, partitionDevice)
    else
        debug(string.format("setPartitionState: Setting non-existant partition %s, variable %s, data %s",
            partition, variable, (data or "N/A")))
    end
end


--
-- Decode the broadcast Label messages for Zones and Partitions
-- TODO: For future use.
--
local function setPanelLabel(data)
    local code, label = data:match("^(%d%d%d)(.*)$")

    code = tonumber(code)
    label = trimSpaces(label)

    if (code &gt;= 1 and code &lt;= 64) then
        debug("setPanelLabel: Zone Label " .. code .. ":" .. label)
        local zoneDevice = g_childDevices.zones[code]

        if (zoneDevice ~= nil) then
            setMiOSAttribute("name", label, zoneDevice)
        end
    elseif (code == 65) then
        debug("setPanelLabel Not set: Fire Alarm Label " .. code .. ":" .. label)
    elseif (code == 66) then
        debug("setPanelLabel Not set: Failed to Arm Label Label " .. code .. ":" .. label)
    elseif (code == 67) then
        debug("setPanelLabel Not set: Alarm when Armed Label " .. code .. ":" .. label)
    elseif (code &gt;= 101 and code &lt;= 108) then
        code = code - 100
        debug("setPanelLabel: Partition Label " .. code .. ":" .. label)
        local partitionDevice = g_childDevices.partitions[code]

        if (partitionDevice ~= nil) then
            setMiOSAttribute("name", label, partitionDevice)
        end

        --
        -- Gross hack, apologies to my CS101 teachers.  I need a spot when I "know" that all the state
        -- establishing calls (Zone Status, Partition Status) have "come through" during startup, so
        -- we tack this on here to clear out the startup bit...
        --
        g_isStartupMode = false
    elseif (code &gt;= 120 and code &lt;= 151) then
        debug("setPanelLabel Not set: Command Output Label " .. code .. ":" .. label)
    else
        debug("setPanelLabel Not set: Unknown code " .. code .. ":" ..label)
    end
end

local function setZoneState(zone, value)
    local zonedev = g_childDevices.zones[zone]

    --
    -- This can happen during startup, for example, since it sends out ALL
    -- potential Zones, whether you have them or not.
    --
    if (not zonedev) then
        debug(string.format("Warning: No device for Zone %d", zone))
        return nil, "No such Zone device"
    end

    if (value == "1") then
        debug(string.format("Zone %d Open", zone))
    else
        debug(string.format("Zone %d Restored", zone))
    end

    luup.variable_set(SECURITY_SID, "Tripped", value, zonedev)
end

local PANEL_LED_STATUS = {
    ["1"] = "Ready",
    ["2"] = "Armed",
    ["3"] = "Memory",
    ["4"] = "Bypass",
    ["5"] = "Trouble",
    ["6"] = "Program",
    ["7"] = "Fire",
    ["8"] = "Backlight",
    ["9"] = "AC"
}

local PANEL_LED_MODES = {
    ["0"] = "OFF",
    ["1"] = "ON",
    ["2"] = "Flashing"
}

local PANEL_ERRORS = {
    ["017"] = "Keybus Busy - Installer Mode",
    ["021"] = "Requested Partition is out of Range",
    ["023"] = "Partition is Not Armed",
    ["024"] = "Partition is Not Ready to Arm",
    ["026"] = "User Code Not Required",
    ["028"] = "Virtual Keypad is Disabled",
    ["029"] = "Not Valid Parameter",
    ["030"] = "Keypad Does Not Come Out of Blank Mode",
    ["031"] = "IT-100 is already in Thermostat menu",
    ["032"] = "IT-100 is Not in Thermostat menu",
    ["033"] = "No Response from Thermostat (or Escort Module)"
}

-- NOTE: IT100 Programmer's guide is incorrect here.  It shows these as being "3-digit" when
--       they're really 1-digit.
--
local PARTITION_ARMED_MODES = {
    ["0"] = "Away",
    ["1"] = "Stay",
    ["2"] = "Away, No Delay",
    ["3"] = "Stay, No Delay",
    ["4"] = "Night"
}


--[[
local PANEL_EVENTS = {
    ["501"] = {"panel", "Command Error",
            function (cmd, zone)
            debug("Panel::Command Error")
        end},
    ["502"] = {"panel", "System Error",
            function (cmd, data)
            debug(string.format("Panel::System Error (%s)",
                                (PANEL_ERRORS[data] or "wtf")))
        end},
    ["570"] = {"panel", "Broadcast Labels",
            function (cmd, data)
            debug(string.format("Panel::Broadcast Labels (%s)", (data or "N/A")))
            setPanelLabel(data)
        end},
    ["601"] = {"partition-zone", "Zone Alarm",
            function (cmd, partition, zone)
            debug(string.format("Zone::Zone Alarm, Partition %d, Zone %d", partition, zone))
        end},
    ["602"] = {"partition-zone", "Zone Alarm Restore",
            function (cmd, partition, zone)
            debug(string.format("Zone::Zone Alarm Restore, Partition %d, Zone %d", partition, zone))
        end},
    ["603"] = {"partition-zone", "Zone Tamper",
            function (cmd, partition, zone)
            debug(string.format("Partition-Zone::Tamper not implemented (%s, %s)", partition, zone))
        end},
    ["604"] = {"partition-zone", "Zone Tamper Restore",
            function (cmd, partition, zone)
            debug(string.format("Partition-Zone::Tamper Restore not implemented (%s, %s)", partition, zone))
        end},
    ["605"] = {"zone", "Zone Fault",
            function (cmd, zone)
            debug(string.format("Zone::Zone Fault not implemented (%s)", zone))
        end},
    ["606"] = {"zone", "Zone Fault Restore",
            function (cmd, zone)
            debug(string.format("Zone::Zone Fault Restore not implemented (%s)", zone))
        end},
    ["609"] = {"zone", "Zone %s Open",
            function (cmd, zone)    -- zone open
            setZoneState(zone, "1")
        end},
    ["610"] = {"zone", "Zone %s Restored",
            function (cmd, zone)    -- zone restored
            setZoneState(zone, "0")
        end},
    ["621"] = {"panel", "Fire Key Alarm Set",
            function (cmd, data)    -- fire key alarm
            setDSCPanelState("FireKeyAlarm", "1")
            task("Fire Key Alarm Set", TASK_ERROR_PERM)
        end},
    ["622"] = {"panel", "Fire Key Alarm Restore",
            function (cmd, data)    -- fire key alarm restored
            setDSCPanelState("FireKeyAlarm", "0")
            task("Fire Key Alarm Restore", TASK_SUCCESS)
        end},
    ["623"] = {"panel", "Aux Key Alarm Set",
            function (cmd, data)    -- aux key alarm
            setDSCPanelState("AuxKeyAlarm", "1")
            task("Aux Key Alarm Set", TASK_ERROR_PERM)
        end},
    ["624"] = {"panel", "Aux Key Alarm Restore",
            function (cmd, data)    -- aux key alarm restored
            setDSCPanelState("AuxKeyAlarm", "0")
            task("Aux Key Alarm Restore", TASK_SUCCESS)
        end},
    ["625"] = {"panel", "Panic Key Alarm Set",
            function (cmd, data)    -- panic key alarm
            setDSCPanelState("PanicKeyAlarm", "1")
            task("Panic Key Alarm Set", TASK_ERROR_PERM)
        end},
    ["626"] = {"panel", "Panic Key Alarm Restore",
            function (cmd, data)    -- panic key alarm restored
            setDSCPanelState("PanicKeyAlarm", "0")
            task("Panic Key Alarm Restore", TASK_SUCCESS)
        end},
    ["631"] = {"panel", "Aux Input Alarm Set",
            function (cmd, data)    -- Aux input alarm
            setDSCPanelState("AuxInputAlarm", "1")
            task("Aux Input Alarm Set", TASK_ERROR_PERM)
        end},
    ["632"] = {"panel", "Aux Input Alarm Restore",
            function (cmd, data)    -- Aux input alarm restored
            setDSCPanelState("AuxInputAlarm", "0")
            task("Aux Input Alarm Restore", TASK_SUCCESS)
        end},
    ["650"] = {"partition", "Ready",
            function (cmd, partition, data)
            debug("Partition::Ready " .. cmd .. " partition: " .. partition)

            setPartitionState(partition, "ArmMode", "Disarmed")
            setPartitionState(partition, "DetailedArmMode", "Ready")

            --
            -- This happens at a few points, but notably during startup, so we need to "clear" all
            -- prior state.
            --
            if (g_isStartupMode) then
                setPartitionState(partition, "ArmMode", "Disarmed")
            end
        end},
    ["651"] = {"partition", "Not Ready",
            function (cmd, partition, data)
            debug("Partition::Not Ready " .. cmd .. " partition: " .. partition)

            setPartitionState(partition, "ArmMode", "Disarmed")
            setPartitionState(partition, "DetailedArmMode", "NotReady")

            --
            -- This happens at a few points, but notably during startup, so we need to "clear" all
            -- prior state.
            --
            if (g_isStartupMode) then
                setPartitionState(partition, "ArmMode", "Disarmed")
            end
        end},
    ["652"] = {"partition", "Partition Armed - Descriptive Mode",
            function (cmd, partition, data)
            debug(string.format("Partition::Partition Armed - Descriptive Mode %s, Partition %d (%s)",
                cmd, partition, (PARTITION_ARMED_MODES[data] or "N/A")))

            setPartitionState(partition, "ArmMode", "Armed")

            if (data=="0") then
                setPartitionState(partition, "DetailedArmMode", "Armed")
            elseif (data=="1") then
                setPartitionState(partition, "DetailedArmMode", "Stay")
            elseif (data=="2") then
                setPartitionState(partition, "DetailedArmMode", "Vacation")
            elseif (data=="3") then
                setPartitionState(partition, "DetailedArmMode", "StayInstant")
            elseif (data=="4") then
                setPartitionState(partition, "DetailedArmMode", "Night")
            else
                log("Received invalid partition state " .. data)
                setPartitionState(partition, "DetailedArmMode", "Armed")
            end
        end},
    ["654"] = {"partition", "Alarm",
            function (cmd, partition, data)
            debug("Partition::Alarm " .. cmd .. " partition: " .. partition)

            setPartitionState(partition, "Alarm", "Active")
            setPartitionState(partition, "AlarmMemory", "1")
            setPartitionState(partition, "LastAlarmActive", os.date())

            if (g_isStartupMode) then
                setPartitionState(partition, "ArmMode", "Armed")
                setPartitionState(partition, "DetailedArmMode", "Armed")
            end
        end},
    ["655"] = {"partition", "Disarmed",
            function (cmd, partition, data)
            debug("Partition::Disarmed " .. cmd .. " partition: " .. partition)

            setPartitionState(partition, "ArmMode", "Disarmed")
            setPartitionState(partition, "DetailedArmMode", "Disarmed")
            setPartitionState(partition, "Alarm", "None")
            setPartitionState(partition, "AlarmMemory", "0")
            setPartitionState(partition, "LastAlarmActive", "0")
        end},
    ["656"] = {"partition", "Exit Delay in Progress",
            function (cmd, partition, data)
            debug("Partition::Exit Delay in Progress " .. cmd .. " partition: " .. partition)

            setPartitionState(partition, "ArmMode", "Disarmed")
            setPartitionState(partition, "DetailedArmMode", "ExitDelay")

            if (g_isStartupMode) then
                setPartitionState(partition, "Alarm", "None")
            end
        end},
    ["657"] = {"partition", "Entry Delay in Progress",
            function (cmd, partition, data)
            debug("Partition::Entry Delay in Progress " .. cmd .. " partition: " .. partition)

            setPartitionState(partition, "ArmMode", "Armed")
            setPartitionState(partition, "DetailedArmMode", "EntryDelay")

            if (g_isStartupMode) then
                setPartitionState(partition, "Alarm", "None")
            end
        end},
    ["658"] = {"partition", "Keypad Lockout",
            function (cmd, partition, data)
            debug("Partition::Keypad Lockout " .. cmd .. " partition: " .. partition)
        end},
    ["659"] = {"partition", "Keypad Blanking",
            function (cmd, partition, data)
            debug("Partition::Keypad Blanking " .. cmd .. " partition: " .. partition)
        end},
    ["660"] = {"partition", "Command Output In Progress",
            function (cmd, partition, data)
            debug("Partition::Command Output In Progress " .. cmd .. " partition: " .. partition)
        end},
    ["670"] = {"partition", "Invalid Access Code",
            function (cmd, partition, data)
            debug("Partition::Invalid Access Code " .. cmd .. " partition: " .. partition)
        end},
    ["671"] = {"partition", "Function Not Available",
            function (cmd, partition, data)
            debug("Partition::Function Not Available " .. cmd .. " partition: " .. partition)
        end},
    ["672"] = {"partition", "Failed To Arm",
            function (cmd, partition, data)
            debug("Partition::Failed To Arm " .. cmd .. " partition: " .. partition)

            setPartitionState(partition, "ArmMode", "Disarmed")
            setPartitionState(partition, "DetailedArmMode", "FailedToArm")

            if (g_isStartupMode) then
                setPartitionState(partition, "Alarm", "None")
            end

            task("Failed to Arm Partition", TASK_ERROR)
        end},
    ["673"] = {"partition", "Partition Busy",
            function (cmd, partition, data)
            debug("Partition::Partition Busy " .. cmd .. " partition: " .. partition)
        end},
    ["700"] = {"partition", "User Closing",
            function (cmd, partition, data)
            debug("Partition::User Closing " .. cmd .. " partition: " .. partition)
        end},
    ["701"] = {"partition", "Special Closing",
            function (cmd, partition, data)
            debug("Partition::Special Closing " .. cmd .. " partition: " .. partition)
        end},
    ["702"] = {"partition", "Partial Closing",
            function (cmd, partition, data)
            debug("Partition::Partial Closing " .. cmd .. " partition: " .. partition)
        end},
    ["750"] = {"partition", "User Opening",
            function (cmd, partition, data)
            debug("Partition::User Opening " .. cmd .. " partition: " .. partition)
        end},
    ["751"] = {"partition", "Special Opening",
            function (cmd, partition, data)
            debug("Partition::Special Opening " .. cmd .. " partition: " .. partition)
        end},
    ["800"] = {"panel", "Panel Battery Trouble",
            function (cmd, data)
            debug("Panel::Panel Battery Trouble " .. cmd)
            task("Panel Battery Trouble", TASK_ERROR_PERM)
        end},
    ["801"] = {"panel", "Panel Battery Restore",
            function (cmd, data)
            debug("Panel::Panel Battery Restore " .. cmd)
            task("Panel Battery Restore", TASK_SUCCESS)
        end},
    ["802"] = {"panel", "Panel AC Trouble",
            function (cmd, data)
            debug("Panel::Panel AC Trouble " .. cmd)
            task("Panel AC Trouble", TASK_ERROR_PERM)
        end},
    ["803"] = {"panel", "Panel AC Restore",
            function (cmd, data)
            debug("Panel::Panel AC Restore " .. cmd)
            task("Panel AC Restore", TASK_SUCCESS)
        end},
    ["806"] = {"panel", "System Bell Trouble",
            function (cmd, data)
            debug("Panel::System Bell Trouble " .. cmd)
            task("System Bell Trouble", TASK_ERROR_PERM)
        end},
    ["807"] = {"panel", "System Bell Restore",
            function (cmd, data)
            debug("System Bell Restore " .. cmd)
            task("System Bell Restore", TASK_SUCCESS)
        end},
    ["829"] = {"panel", "General System Tamper",
            function (cmd, data)
            debug("Panel::General System Tamper " .. cmd)
            task("System General System Tamper", TASK_ERROR_PERM)
        end},
    ["830"] = {"panel", "General System Tamper Restore",
            function (cmd, data)
            debug("General System Tamper Restore" .. cmd)
            task("General System Tamper Restore", TASK_SUCCESS)
        end},
    ["840"] = {"partition", "Trouble Status (LED ON)",
            function (cmd, partition, data)
            debug("Partition::Trouble Status (LED ON) " .. cmd .. " partition: " .. partition)
        end},
    ["841"] = {"partition", "Trouble Status (LED OFF)",
            function (cmd, partition, data)
            debug("Partition::Trouble Status (LED OFF) " .. cmd .. " partition: " .. partition)
        end},
    ["842"] = {"panel", "Fire Trouble Alarm",
            function (cmd, data)
            debug("Panel::Fire Trouble Alarm " .. cmd)
            task("Fire Trouble Alarm", TASK_ERROR_PERM)
        end},
    ["843"] = {"panel", "Fire Trouble Alarm Restored",
            function (cmd, data)
            debug("Fire Trouble Alarm Restored " .. cmd)
            task("Fire Trouble Alarm Restored", TASK_SUCCESS)
        end},
    ["896"] = {"panel", "Keybus Fault",
            function (cmd, data)
            debug("Panel::Keybus Fault " .. cmd)
            task("Keybus Fault", TASK_ERROR_PERM)
        end},
    ["897"] = {"panel", "Keybus Fault Restored",
            function (cmd, data)
            debug("Panel::Keybus Fault Restored " .. cmd)
            task("Keybus Fault Restored", TASK_SUCCESS)
        end},
    ["900"] = {"partition", "Code Required",
            function (cmd, partition, data)
            debug("Partition::Code Required " .. cmd .. " partition: " .. partition)
        end},
    ["901"] = {"panel", "LCD Update",
            function (cmd, data)    -- LCD Update
            -- Strip out the first 6 characters (formatting/layout)
            debug("Panel::LCD Update " .. cmd .. data)
            setDSCPanelState("LCD", trimSpaces(data:sub(6)))
        end},
    ["903"] = {"panel", "LED Status",
            function (cmd, data)
            debug(string.format("Panel::LED Status %s (%s)",
                                (PANEL_LED_STATUS[data:sub(1,1)] or "wtf"),
                                 (PANEL_LED_MODES[data:sub(2,2)] or "wtf")))
        end},
    ["904"] = {"panel", "Beep Status",
            function (cmd, data)
            debug("Panel::Beep Status " .. cmd .. " data: " .. data)
        end},
    ["905"] = {"panel", "Tone Status",
            function (cmd, data)
            debug("Panel::Tone Status " .. cmd .. " data: " .. data)
        end},
    ["906"] = {"panel", "Buzzer Status",
            function (cmd, data)
            debug("Panel::Buzzer Status " .. cmd .. " data: " .. data)
        end},
    ["908"] = {"panel", "Software Version",
            function (cmd, data)
            debug("Panel::Software Version " .. cmd .. " data: " .. data)
            setDSCPanelState("FirmwareVersion", data)
        end}
}]]

local ZONE_DEFINITIONS = {
    ['0'] = {"Disabled"},
    ['1'] = {"Burglar Entry/Exit 1"},
    ['2'] = {"Burglar Entry/Exit 2"},
    ['3'] = {"Burglar Perimeter Instant"},
    ['4'] = {"Burglar Interior"},
    ['5'] = {"Burglar Interior Follower"},
    ['6'] = {"Burglar Interior Night"},
    ['7'] = {"Burglar Interior Night Delay"},
    ['8'] = {"Burglar 24 Hour"},
    ['9'] = {"Burglar Box Tamper"},
    [':'] = {"Fire Alarm"},
    [';'] = {"Fire Verified"},
    ['<'] = {"Fire Supervisory"},
    ['='] = {"Aux Alarm 1"},
    ['>'] = {"Aux Alarm 2"},
    ['?'] = {"Keyfob"},
    ['@'] = {"Non Alarm"},
    ['A'] = {"Carbon Monoxide"},
    ['B'] = {"Emergency Alarm"},
    ['C'] = {"Freeze Alarm"},
    ['D'] = {"Gas Alarm"},
    ['E'] = {"Heat Alarm"},
    ['F'] = {"Medical Alarm"},
    ['G'] = {"Police Alarm"},
    ['H'] = {"Police No Indication"},
    ['I'] = {"Water Alarm"},
    ['J'] = {"Key Momentary Arm / Disarm"},
    ['K'] = {"Key Momentary Arm Away"},
    ['L'] = {"Key Momentary Arm Stay"},
    ['M'] = {"Key Momentary Disarm"},
    ['N'] = {"Key On/Off"},
    ['O'] = {"Mute Audibles"},
    ['P'] = {"Power Supervisory"},
    ['Q'] = {"Temperature"},
    ['R'] = {"Analog Zone"},
    ['S'] = {"Phone Key"},
    ['T'] = {"Intercom Key"}
}


local ZONE_STATUSES = {
    ['0'] = {"Normal", "Unconfigured"},
    ['1'] = {"Normal", "Open"},
    ['2'] = {"Normal", "EOL"},
    ['3'] = {"Normal", "Short"},
    ['4'] = nil,
    ['5'] = {"Trouble", "Open"},
    ['6'] = {"Trouble", "EOL"},
    ['7'] = {"Trouble", "Short"},
    ['8'] = nil,
    ['9'] = {"Violated", "Open"},
    ['A'] = {"Violated", "EOL"},
    ['B'] = {"Violated", "Short"},
    ['C'] = {"Bypassed", nil},
    ['D'] = {"Bypassed", "Open"},
    ['E'] = {"Bypassed", "EOL"},
    ['F'] = {"Bypassed", "Short"}
}


local function getZoneDefinition (data)
    for i = 1, #data do
        c = data:sub (i, i)
        if (c ~= '0') then
            g_childDevices.zones[i] = {description = ZONE_DEFINITIONS[c]}
        end
    end
end


local function getZonePartitionReport (data)
    for i = 1, #data do
        c = data:sub (i, i)
        if (g_childDevices.zones[i]) then
            if (c ~= '0')
                g_childDevices.zones[i].partition = c;
                if (not table.contains (g_childDevices.partitions, c)) then
                    table.insert (g_childDevices.partitions, c)
                end
            else
                g_childDevices.zones[i].partition = "no partition";
            end
        end
    end
    table.sort (g_childDevices.partitions)
end


local function getZoneStatusReport (data)
    for i = 1, #data do
        c = data:sub (i, i)
        if (g_childDevices.zones[i]) then
            g_childDevices.zones[i].status = ZONE_STATUSES[c][1]
        end
    end
end


local PANEL_RESPONSES = {
    ["ZD"] = {
        "Zone Definition Data",
        function (data)
            debug ("Message type description: Zone Definition Data")
            getZoneDefinition (data)
        end
    },
    ["ZP"] = {
        "Zone Partition Report",
        function (data)
            debug ("Message type description: Zone Partion Report")
            getZonePartitionReport (data)
        end
    },
    ["ZS"] = {
        "Zone Status Report",
        function (data)
            debug ("Message type description: Zone Status Report")
            getZoneStatusReport (data)
        end
    }
}


local function sendCommand (command, data)
    data = data or ""
    
    local length = 2 + #data + 2 + 2 -- command length + data length + 00 length + checksum length
    local checksum = calculateChecksum (string.format ("%0.2X%s%s00", length, command, data))
    local cmd = string.format ("%0.2X%s%s00%0.2X", length, command, data, checksum)
    if (luup.io.write (cmd) == false) then
        local message = string.format ("ERROR: Cannot send command: '%s'", cmd)
        log (message)
        luup.set_failure (true)

        return false, message
    end

    return true, "Ok"
end


-- Get and return the variable. If the variable is 'nil' it is created and
-- initialized to the given default value.
local function presetVariable (serviceId, variable, device, defaultValue)
    local v = luup.variable_get (serviceId, variable, device)
    if (v == nil) then
        luup.variable_set (serviceId, variable, (defaultValue or ""), device)
    end

    return v
end


local function presetParameters (t)
    local params = {}
    for _, param in ipairs (t) do
        table.insert (params, string.format ("%s,%s=%s", unpack (param)))
    end

    return table.concat (params, "\n")
end


local function appendChildDevices (mainDevice, rootptr, partitions, id, description, deviceType, deviceFilename, parameters)
    for item in partitions:gmatch ("[^,]+") do
        item = trimSpaces (item)
        if (item and item ~= "") then
            debug (string.format ("Appending Child: %s", item))
            luup.chdev.append (mainDevice, rootptr, id:format (item), description:format (item), deviceType, deviceFilename, nil, (parameters or ""), false)
        else
            debug ("Data formatting problem for CSV: " .. partitions)
        end
    end
end


local function initializeParameters (device)
    local enableRemoteArming = luup.variable_get (ELK_PANEL_SID, "EnableRemoteArm", device)
    if (enableRemoteArming == nil) then
        g_remoteArmEnabled = false
        g_remoteDisarmEnabled = false
        luup.variable_set (ELK_PANEL_SID, "EnableRemoteArm", "false", device)
    elseif (enableRemoteArming  == "false") then
        g_remoteArmEnabled = false
        g_remoteDisarmEnabled = false
    elseif (enableRemoteArming == "arm") then
        g_remoteArmEnabled = true
        g_remoteDisarmEnabled = false
    elseif (enableRemoteArming == "disarm") then
        g_remoteArmEnabled = true
        g_remoteDisarmEnabled = true
    else
        return false, "EnableRemoteArm parameter must be either blank, false or a value from the wiki...", "ElkAlarmPanel"
    end

    local enableRemotePanic = luup.variable_get (ELK_PANEL_SID, "EnableRemotePanic", device)
    if (enableRemotePanic == nil) then
        g_remotePanicEnabled = false
        luup.variable_set (ELK_PANEL_SID, "EnableRemotePanic", "false", device)
    elseif (enableRemotePanic == "false") then
        g_remotePanicEnabled = false
    elseif (enableRemotePanic == "panic") then
        g_remotePanicEnabled = true
    else
        return false, "EnableRemotePanic parameter must be either blank, false or a value from the wiki..."
    end

    return true, "Ok"
end


local function initializeLabelsAndAlarmState (device)
    local status, message

    -- Trigger the Panel to emit its State information for the Panel, Partitions and Zones.
    status, message = sendCommand ("001")
    if (status) then
        -- Trigger the Panel to emit it's labels for all Partitions and Zones.
        status, message = sendCommand("002")
    end

    return status, message
end

function sendPartitionCommand(cmd, device, ...)
    local partition = extractPartition(luup.devices[device].id)
    assert(partition ~= nil)

    sendCommand(cmd, partition, ...)
end

----------------------------------------------------------------------

function calculateChecksum (s)
    local total = 0
    for i = 1, #s do
        total = total + string.byte (s, i)
    end

    return 256 - (total % 256)
end


function processEvent (msgType, data)
    local response = PANEL_RESPONSES[msgType]

    if (response == nil) then
        debug ("Unhandled event " .. msgType)
        return
    end

    local eventDescription, eventFunction = response[1], response[2]

    if (type (eventFunction) ~= "function") then
        log ("ERROR: Unknown event, or event handled incorrectly.")
        return
    end

    --[[if (eventType == "zone") then
        return eventFunction(cmd, tonumber(data))
    elseif (eventType == "panel") then
        return eventFunction(cmd, data)
    elseif (eventType == "partition") then
        local partition, data = data:match("^([1-8])(.*)$")
        partition = tonumber(partition)

        setPartitionState(partition, "VendorStatusCode", cmd)
        setPartitionState(partition, "VendorStatus", eventDescription)

        return eventFunction(cmd, partition, data)
    elseif (eventType == "partition-zone") then
        local partition, zone = data:match("^(%d)(%d%d%d)")
        return eventFunction(cmd, tonumber(partition), tonumber(zone))
    else
        debug("wtf? :" .. cmd .. "(" .. eventType .. ")")
    end]]
end


-- If the input line is a response to a pending command, resumes it.
-- If it's an error, handles it.
-- If it's an event, calls handler.
--local pending = nil
function processIncoming (s)
    local length, msgType, data, checksum = s:match ("^(%x%x)(%a%w)(.*)00(%x%x)$")
    debug (string.format ("processIncoming: length=%s, message type=%s, data='%s', checksum=%s", length, msgType, data, checksum))
    
    --
    -- Validate received message.
    --
    local calculatedLength = #msgType + #data + 2 + #checksum
    if (calculatedLength ~= tonumber (length, 16)) then
        log ("Data transmission error: the received message length is incorrect.")
        task ("Data transmission error.", TASK_ERROR)
        return
    end
    local calculatedChecksum = calculateChecksum (length .. msgType .. data .. "00")
    if (calculatedChecksum ~= tonumber (checksum, 16)) then
        log ("Data transmission error: the calculated checksum is incorrect.")
        task ("Data transmission error.", TASK_ERROR)
        return
    end
    --[[if (msgType == "500" or cmd == "900") then
        if (type(pending)=="function") then
            debug("continue with pending callback")
            return pending(cmd, data)
        end
    else
        return processEvent(cmd, data)
    end]]
    return processEvent (msgType, data)
end

function commandCallback(cmd, data, cb)
    pending = nil

    local status, message = sendCommand(cmd, data)
    if (status) then
        debug(string.format("sent (%s,%s), setup callback (%s)", cmd, data, tostring(cb)))
        pending = cb
    end
end

function command(cmd, data, cb)
    return commandCallback(cmd, data, function(rcmd, rdata)
        if (rcmd == "500" and rdata == cmd) then
            return cb and cb(rcmd, rdata) or true
        end
    end)
end

function commandAccess(cmd, data, accesscode, cb)
    return commandCallback(cmd, data, function(rcmd, rdata)
        if (rcmd == "500" and rdata == cmd) then
            return cb and cb(rcmd, rdata) or true
        elseif (rcmd == "900") then
            return commandCallback("200", accesscode, function(rcmd, rdata)
                if (rcmd == "500" and rdata == cmd) then
                    return cb and cb(rcmd, rdata) or true
                end
            end)
        end
    end)
end

--------------------------------------------------------------------------------------------------------------

local function extractPartition (id)
    local partition = id:match ("^elk_partition_(%d+)$")
    debug ("extractPartition: Partition " .. (partition or "N/A"))

    return partition
end


local function extractZone (id)
    local zone = id:match ("^elk_zone_(%d+)$")
    debug("extractZone:Zone " .. (zone or "nil"))

    return zone
end


local function initializeChildren()
    g_childDevices.partitions = {}
    
    for k, v in pairs (luup.devices) do
        if (v.device_num_parent == g_alarmDevice) then
            debug (string.format ("Scanning luup.devices[%s(%s)]", tostring (k), type (k)))
            local zone = extractZone (v.id)
            if (zone ~= nil) then
                g_childDevices.zones[tonumber (zone)].device = k
            else
                local partition = extractPartition (v.id)
                if (partition ~= nil) then
                    g_childDevices.partitions[tonumber (partition)] = k
                else
                    debug (string.format ("Found unexpected child: device #%d, id=%s", k, v.id))
                end
            end
        end
    end

    return true, "Ok"
end


local function getZoneParams (zoneNo)
    local armed
    if ((not g_childDevices.zones[zoneNo].status) or (g_childDevices.zones[zoneNo].status == "Bypassed")) then
        armed = "0"
    else
        armed = "1"
    end
    
    return presetParameters ({{SECURITY_SID, "Armed", armed}})
end


local function getDeviceInfo (zoneNo)
    -- TODO: the actual implementation
    return {type = "urn:schemas-micasaverde-com:device:MotionSensor:1", file = "D_MotionSensor1.xml"}
end


local function appendZones (rootPtr)
    for k, v in pairs (g_childDevices.zones) do
        debug (string.format ("Appending Zone %d", k))
        local device = getDeviceInfo (k)
        luup.chdev.append (g_alarmDevice, rootPtr, string.format ("elk_zone_%d", k), string.format ("Zone %d: %s", k, v.description), device.type, device.file, nil, getZoneParams (k), false)
    end
end


local function appendPartitions (rootPtr)
    for i = 1, #g_childDevices.partitions do
        local p = g_childDevices.partitions[i]
        debug (string.format ("Appending Partition %c", p))
        luup.chdev.append (g_alarmDevice, rootPtr, string.format ("elk_partition_%c", p), string.format ("Partition %c", p), "urn:schemas-micasaverde-com:device:ElkAlarmPartition:1", "D_ElkAlarmPartition1.xml")
    end
end


local function getZonesAndPartitions()
    -- Get zone information.
    local status, message = sendCommand ("zd")
    if (status) then
        -- Get partition information.
        status, message = sendCommand ("zp")
    end
    
    return status, message
end


function alarmStartup (lul_device)
    log ("Elk M1 Alarm Panel Plugin ver. " .. VERSION .." (ID = " .. tostring (lul_device) .. ")")
    
    g_alarmDevice = lul_device
    
    local rootPtr = luup.chdev.start (lul_device)

    -- Get all the zones and active partitions.
    local status, message = getZonesAndPartitions()
    if (not status) then
        return status, message, "Elk Alarm Panel"
    end
    
    -- Create a device for each active partition.
    appendPartitions (rootPtr)
    
    -- Get each Zone's status.
    local status, message = sendCommand("zs")
    
    -- Create a device for each zone.
    appendZones (rootPtr)
    
    luup.chdev.sync (lul_device, rootPtr)

    local status, message = initializeParameters (lul_device)
    if (status) then
        status, message = initializeChildren (lul_device)
        if (status) then
            status, message = initializeLabelsAndAlarmState (lul_device)
        end
    end

    return status, message, "Elk Alarm Panel"
end